@@ -1,4 +1,4 @@
-# Copyright 2022 The Nerfstudio Team. All rights reserved.
+# Copyright 2022 the Regents of the University of California, Nerfstudio Team and contributors. All rights reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -17,20 +17,14 @@ import math
 import os
 from dataclasses import dataclass, field
 from pathlib import Path
-from typing import Optional, Tuple, Type
+from typing import Literal, Optional, Tuple, Type
 
 import numpy as np
 import pyquaternion
 import torch
-from nuscenes.nuscenes import NuScenes as NuScenesDatabase
-from typing_extensions import Literal
 
 from nerfstudio.cameras.cameras import Cameras, CameraType
-from nerfstudio.data.dataparsers.base_dataparser import (
-    DataParser,
-    DataParserConfig,
-    DataparserOutputs,
-)
+from nerfstudio.data.dataparsers.base_dataparser import DataParser, DataParserConfig, DataparserOutputs
 from nerfstudio.data.scene_box import SceneBox
 
 
@@ -56,7 +50,7 @@ class NuScenesDataParserConfig(DataParserConfig):
     Each clip was recorded with a suite of sensors including 6 surround cameras.
     It also includes 3D cuboid annotations around objects.
     We optionally use these cuboids to mask dynamic objects by specifying the mask_dir flag.
-    To create these masks use scripts/datasets/process_nuscenes_masks.py.
+    To create these masks use nerfstudio/scripts/datasets/process_nuscenes_masks.py.
     """
 
     _target: Type = field(default_factory=lambda: NuScenes)
@@ -86,9 +80,14 @@ class NuScenes(DataParser):
     config: NuScenesDataParserConfig
 
     def _generate_dataparser_outputs(self, split="train"):
-        # pylint: disable=too-many-statements
+        # nuscenes is slow to import, so we only do it if we need it.
+        from nuscenes.nuscenes import NuScenes as NuScenesDatabase
 
-        nusc = NuScenesDatabase(version=self.config.version, dataroot=self.config.data_dir, verbose=self.config.verbose)
+        nusc = NuScenesDatabase(
+            version=self.config.version,
+            dataroot=str(self.config.data_dir.absolute()),
+            verbose=self.config.verbose,
+        )
         cameras = ["CAM_" + camera for camera in self.config.cameras]
 
         assert (
@@ -199,10 +198,10 @@ class NuScenes(DataParser):
         )
 
         cameras = Cameras(
-            fx=intrinsics[:, 0, 0],
-            fy=intrinsics[:, 1, 1],
-            cx=intrinsics[:, 0, 2],
-            cy=intrinsics[:, 1, 2],
+            fx=intrinsics[:, 0, 0].detach().clone(),
+            fy=intrinsics[:, 1, 1].detach().clone(),
+            cx=intrinsics[:, 0, 2].detach().clone(),
+            cy=intrinsics[:, 1, 2].detach().clone(),
             height=900,
             width=1600,
             camera_to_worlds=poses[:, :3, :4],
