@@ -1,4 +1,4 @@
-# Copyright 2022 The Nerfstudio Team. All rights reserved.
+# Copyright 2022 the Regents of the University of California, Nerfstudio Team and contributors. All rights reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -16,14 +16,15 @@
 Collection of Losses.
 """
 from enum import Enum
+from typing import Dict, Literal, Optional, Tuple, cast
 
 import torch
-from torch import nn
-from torchtyping import TensorType
-from typing_extensions import Literal
+from jaxtyping import Bool, Float
+from torch import Tensor, nn
 
 from nerfstudio.cameras.rays import RaySamples
-from nerfstudio.utils.math import masked_reduction
+from nerfstudio.field_components.field_heads import FieldHeadNames
+from nerfstudio.utils.math import masked_reduction, normalized_depth_scale_and_shift
 
 L1Loss = nn.L1Loss
 MSELoss = nn.MSELoss
@@ -41,15 +42,20 @@ class DepthLossType(Enum):
 
     DS_NERF = 1
     URF = 2
+    SPARSENERF_RANKING = 3
+
+
+FORCE_PSEUDODEPTH_LOSS = False
+PSEUDODEPTH_COMPATIBLE_LOSSES = (DepthLossType.SPARSENERF_RANKING,)
 
 
 def outer(
-    t0_starts: TensorType[..., "num_samples_0"],
-    t0_ends: TensorType[..., "num_samples_0"],
-    t1_starts: TensorType[..., "num_samples_1"],
-    t1_ends: TensorType[..., "num_samples_1"],
-    y1: TensorType[..., "num_samples_1"],
-) -> TensorType[..., "num_samples_0"]:
+    t0_starts: Float[Tensor, "*batch num_samples_0"],
+    t0_ends: Float[Tensor, "*batch num_samples_0"],
+    t1_starts: Float[Tensor, "*batch num_samples_1"],
+    t1_ends: Float[Tensor, "*batch num_samples_1"],
+    y1: Float[Tensor, "*batch num_samples_1"],
+) -> Float[Tensor, "*batch num_samples_0"]:
     """Faster version of
 
     https://github.com/kakaobrain/NeRF-Factory/blob/f61bb8744a5cb4820a4d968fb3bfbed777550f4a/src/model/mipnerf360/helper.py#L117
@@ -76,10 +82,10 @@ def outer(
 
 
 def lossfun_outer(
-    t: TensorType[..., "num_samples+1"],
-    w: TensorType[..., "num_samples"],
-    t_env: TensorType[..., "num_samples+1"],
-    w_env: TensorType[..., "num_samples"],
+    t: Float[Tensor, "*batch num_samples_1"],
+    w: Float[Tensor, "*batch num_samples"],
+    t_env: Float[Tensor, "*batch num_samples_1"],
+    w_env: Float[Tensor, "*batch num_samples"],
 ):
     """
     https://github.com/kakaobrain/NeRF-Factory/blob/f61bb8744a5cb4820a4d968fb3bfbed777550f4a/src/model/mipnerf360/helper.py#L136
@@ -103,7 +109,7 @@ def ray_samples_to_sdist(ray_samples):
     return sdist
 
 
-def interlevel_loss(weights_list, ray_samples_list):
+def interlevel_loss(weights_list, ray_samples_list) -> torch.Tensor:
     """Calculates the proposal loss in the MipNeRF-360 paper.
 
     https://github.com/kakaobrain/NeRF-Factory/blob/f61bb8744a5cb4820a4d968fb3bfbed777550f4a/src/model/mipnerf360/model.py#L515
@@ -111,12 +117,16 @@ def interlevel_loss(weights_list, ray_samples_list):
     """
     c = ray_samples_to_sdist(ray_samples_list[-1]).detach()
     w = weights_list[-1][..., 0].detach()
+    assert len(ray_samples_list) > 0
+
     loss_interlevel = 0.0
     for ray_samples, weights in zip(ray_samples_list[:-1], weights_list[:-1]):
         sdist = ray_samples_to_sdist(ray_samples)
         cp = sdist  # (num_rays, num_samples + 1)
         wp = weights[..., 0]  # (num_rays, num_samples)
         loss_interlevel += torch.mean(lossfun_outer(c, w, cp, wp))
+
+    assert isinstance(loss_interlevel, Tensor)
     return loss_interlevel
 
 
@@ -145,9 +155,9 @@ def distortion_loss(weights_list, ray_samples_list):
 
 def nerfstudio_distortion_loss(
     ray_samples: RaySamples,
-    densities: TensorType["bs":..., "num_samples", 1] = None,
-    weights: TensorType["bs":..., "num_samples", 1] = None,
-) -> TensorType["bs":..., 1]:
+    densities: Optional[Float[Tensor, "*bs num_samples 1"]] = None,
+    weights: Optional[Float[Tensor, "*bs num_samples 1"]] = None,
+) -> Float[Tensor, "*bs 1"]:
     """Ray based distortion loss proposed in MipNeRF-360. Returns distortion Loss.
 
     .. math::
@@ -165,10 +175,12 @@ def nerfstudio_distortion_loss(
     """
     if torch.is_tensor(densities):
         assert not torch.is_tensor(weights), "Cannot use both densities and weights"
+        assert densities is not None
         # Compute the weight at each sample location
         weights = ray_samples.get_weights(densities)
     if torch.is_tensor(weights):
         assert not torch.is_tensor(densities), "Cannot use both densities and weights"
+    assert weights is not None
 
     starts = ray_samples.spacing_starts
     ends = ray_samples.spacing_ends
@@ -186,9 +198,9 @@ def nerfstudio_distortion_loss(
 
 
 def orientation_loss(
-    weights: TensorType["bs":..., "num_samples", 1],
-    normals: TensorType["bs":..., "num_samples", 3],
-    viewdirs: TensorType["bs":..., 3],
+    weights: Float[Tensor, "*bs num_samples 1"],
+    normals: Float[Tensor, "*bs num_samples 3"],
+    viewdirs: Float[Tensor, "*bs 3"],
 ):
     """Orientation loss proposed in Ref-NeRF.
     Loss that encourages that all visible normals are facing towards the camera.
@@ -196,26 +208,26 @@ def orientation_loss(
     w = weights
     n = normals
     v = viewdirs * -1
-    n_dot_v = (n * v[..., None, :]).sum(axis=-1)
+    n_dot_v = (n * v[..., None, :]).sum(dim=-1)
     return (w[..., 0] * torch.fmin(torch.zeros_like(n_dot_v), n_dot_v) ** 2).sum(dim=-1)
 
 
 def pred_normal_loss(
-    weights: TensorType["bs":..., "num_samples", 1],
-    normals: TensorType["bs":..., "num_samples", 3],
-    pred_normals: TensorType["bs":..., "num_samples", 3],
+    weights: Float[Tensor, "*bs num_samples 1"],
+    normals: Float[Tensor, "*bs num_samples 3"],
+    pred_normals: Float[Tensor, "*bs num_samples 3"],
 ):
     """Loss between normals calculated from density and normals from prediction network."""
     return (weights[..., 0] * (1.0 - torch.sum(normals * pred_normals, dim=-1))).sum(dim=-1)
 
 
 def ds_nerf_depth_loss(
-    weights: TensorType[..., "num_samples", 1],
-    termination_depth: TensorType[..., 1],
-    steps: TensorType[..., "num_samples", 1],
-    lengths: TensorType[..., "num_samples", 1],
-    sigma: TensorType[0],
-) -> TensorType[..., 1]:
+    weights: Float[Tensor, "*batch num_samples 1"],
+    termination_depth: Float[Tensor, "*batch 1"],
+    steps: Float[Tensor, "*batch num_samples 1"],
+    lengths: Float[Tensor, "*batch num_samples 1"],
+    sigma: Float[Tensor, "0"],
+) -> Float[Tensor, "*batch 1"]:
     """Depth loss from Depth-supervised NeRF (Deng et al., 2022).
 
     Args:
@@ -235,12 +247,12 @@ def ds_nerf_depth_loss(
 
 
 def urban_radiance_field_depth_loss(
-    weights: TensorType[..., "num_samples", 1],
-    termination_depth: TensorType[..., 1],
-    predicted_depth: TensorType[..., 1],
-    steps: TensorType[..., "num_samples", 1],
-    sigma: TensorType[0],
-) -> TensorType[..., 1]:
+    weights: Float[Tensor, "*batch num_samples 1"],
+    termination_depth: Float[Tensor, "*batch 1"],
+    predicted_depth: Float[Tensor, "*batch 1"],
+    steps: Float[Tensor, "*batch num_samples 1"],
+    sigma: Float[Tensor, "0"],
+) -> Float[Tensor, "*batch 1"]:
     """Lidar losses from Urban Radiance Fields (Rematas et al., 2022).
 
     Args:
@@ -274,15 +286,15 @@ def urban_radiance_field_depth_loss(
 
 
 def depth_loss(
-    weights: TensorType[..., "num_samples", 1],
+    weights: Float[Tensor, "*batch num_samples 1"],
     ray_samples: RaySamples,
-    termination_depth: TensorType[..., 1],
-    predicted_depth: TensorType[..., 1],
-    sigma: TensorType[0],
-    directions_norm: TensorType[..., 1],
+    termination_depth: Float[Tensor, "*batch 1"],
+    predicted_depth: Float[Tensor, "*batch 1"],
+    sigma: Float[Tensor, "0"],
+    directions_norm: Float[Tensor, "*batch 1"],
     is_euclidean: bool,
     depth_loss_type: DepthLossType,
-) -> TensorType[0]:
+) -> Float[Tensor, "0"]:
     """Implementation of depth losses.
 
     Args:
@@ -313,8 +325,8 @@ def depth_loss(
 
 
 def monosdf_normal_loss(
-    normal_pred: TensorType["num_samples", 3], normal_gt: TensorType["num_samples", 3]
-) -> TensorType[0]:
+    normal_pred: Float[Tensor, "num_samples 3"], normal_gt: Float[Tensor, "num_samples 3"]
+) -> Float[Tensor, "0"]:
     """
     Normal consistency loss proposed in monosdf - https://niujinshuchong.github.io/monosdf/
     Enforces consistency between the volume rendered normal and the predicted monocular normal.
@@ -338,13 +350,16 @@ class MiDaSMSELoss(nn.Module):
     def __init__(self, reduction_type: Literal["image", "batch"] = "batch"):
         super().__init__()
 
-        self.reduction_type = reduction_type
+        self.reduction_type: Literal["image", "batch"] = reduction_type
         # reduction here is different from the image/batch-based reduction. This is either "mean" or "sum"
         self.mse_loss = MSELoss(reduction="none")
 
     def forward(
-        self, prediction: TensorType[1, 32, "mult"], target: TensorType[1, 32, "mult"], mask: TensorType[1, 32, "mult"]
-    ) -> TensorType[0]:
+        self,
+        prediction: Float[Tensor, "1 32 mult"],
+        target: Float[Tensor, "1 32 mult"],
+        mask: Bool[Tensor, "1 32 mult"],
+    ) -> Float[Tensor, "0"]:
         """
         Args:
             prediction: predicted depth map
@@ -355,7 +370,8 @@ class MiDaSMSELoss(nn.Module):
         """
         summed_mask = torch.sum(mask, (1, 2))
         image_loss = torch.sum(self.mse_loss(prediction, target) * mask, (1, 2))
-        image_loss = masked_reduction(image_loss, summed_mask, self.reduction_type)
+        # multiply by 2 magic number?
+        image_loss = masked_reduction(image_loss, 2 * summed_mask, self.reduction_type)
 
         return image_loss
 
@@ -375,12 +391,15 @@ class GradientLoss(nn.Module):
             reduction_type: either "batch" or "image"
         """
         super().__init__()
-        self.reduction_type = reduction_type
+        self.reduction_type: Literal["image", "batch"] = reduction_type
         self.__scales = scales
 
     def forward(
-        self, prediction: TensorType[1, 32, "mult"], target: TensorType[1, 32, "mult"], mask: TensorType[1, 32, "mult"]
-    ) -> TensorType[0]:
+        self,
+        prediction: Float[Tensor, "1 32 mult"],
+        target: Float[Tensor, "1 32 mult"],
+        mask: Bool[Tensor, "1 32 mult"],
+    ) -> Float[Tensor, "0"]:
         """
         Args:
             prediction: predicted depth map
@@ -389,7 +408,8 @@ class GradientLoss(nn.Module):
         Returns:
             gradient loss based on reduction function
         """
-        total = 0
+        assert self.__scales >= 1
+        total = 0.0
 
         for scale in range(self.__scales):
             step = pow(2, scale)
@@ -401,11 +421,15 @@ class GradientLoss(nn.Module):
             )
             total += grad_loss
 
+        assert isinstance(total, Tensor)
         return total
 
     def gradient_loss(
-        self, prediction: TensorType[1, 32, "mult"], target: TensorType[1, 32, "mult"], mask: TensorType[1, 32, "mult"]
-    ) -> TensorType[0]:
+        self,
+        prediction: Float[Tensor, "1 32 mult"],
+        target: Float[Tensor, "1 32 mult"],
+        mask: Bool[Tensor, "1 32 mult"],
+    ) -> Float[Tensor, "0"]:
         """
         multiscale, scale-invariant gradient matching term to the disparity space.
         This term biases discontinuities to be sharp and to coincide with discontinuities in the ground truth
@@ -456,50 +480,12 @@ class ScaleAndShiftInvariantLoss(nn.Module):
 
         self.__prediction_ssi = None
 
-    @classmethod
-    def compute_scale_and_shift(
-        cls, prediction: TensorType[1, 32, "mult"], target: TensorType[1, 32, "mult"], mask: TensorType[1, 32, "mult"]
-    ):
-        """
-        More info here: https://arxiv.org/pdf/2206.00665.pdf supplementary section A2 Depth Consistency Loss
-        This function computes scale/shift required to normalizes predicted depth map,
-        to allow for using normalized depth maps as input from monocular depth estimation networks.
-        These networks are trained such that they predict normalized depth maps.
-
-        Solves for scale/shift using a least squares approach with a closed form solution:
-        Based on:
-        https://github.com/autonomousvision/monosdf/blob/d9619e948bf3d85c6adec1a643f679e2e8e84d4b/code/model/loss.py#L7
-        Args:
-            prediction: predicted depth map
-            target: ground truth depth map
-            mask: mask of valid pixels
-        Returns:
-            scale and shift for depth prediction
-        """
-        # system matrix: A = [[a_00, a_01], [a_10, a_11]]
-        a_00 = torch.sum(mask * prediction * prediction, (1, 2))
-        a_01 = torch.sum(mask * prediction, (1, 2))
-        a_11 = torch.sum(mask, (1, 2))
-
-        # right hand side: b = [b_0, b_1]
-        b_0 = torch.sum(mask * prediction * target, (1, 2))
-        b_1 = torch.sum(mask * target, (1, 2))
-
-        # solution: x = A^-1 . b = [[a_11, -a_01], [-a_10, a_00]] / (a_00 * a_11 - a_01 * a_10) . b
-        scale = torch.zeros_like(b_0)
-        shift = torch.zeros_like(b_1)
-
-        det = a_00 * a_11 - a_01 * a_01
-        valid = det.nonzero()
-
-        scale[valid] = (a_11[valid] * b_0[valid] - a_01[valid] * b_1[valid]) / det[valid]
-        shift[valid] = (-a_01[valid] * b_0[valid] + a_00[valid] * b_1[valid]) / det[valid]
-
-        return scale, shift
-
     def forward(
-        self, prediction: TensorType[1, 32, "mult"], target: TensorType[1, 32, "mult"], mask: TensorType[1, 32, "mult"]
-    ) -> TensorType[0]:
+        self,
+        prediction: Float[Tensor, "1 32 mult"],
+        target: Float[Tensor, "1 32 mult"],
+        mask: Bool[Tensor, "1 32 mult"],
+    ) -> Float[Tensor, "0"]:
         """
         Args:
             prediction: predicted depth map (unnormalized)
@@ -508,7 +494,7 @@ class ScaleAndShiftInvariantLoss(nn.Module):
         Returns:
             scale and shift invariant loss
         """
-        scale, shift = ScaleAndShiftInvariantLoss.compute_scale_and_shift(prediction, target, mask)
+        scale, shift = normalized_depth_scale_and_shift(prediction, target, mask)
         self.__prediction_ssi = scale.view(-1, 1, 1) * prediction + shift.view(-1, 1, 1)
 
         total = self.__data_loss(self.__prediction_ssi, target, mask)
@@ -525,3 +511,75 @@ class ScaleAndShiftInvariantLoss(nn.Module):
         return self.__prediction_ssi
 
     prediction_ssi = property(__get_prediction_ssi)
+
+
+def tv_loss(grids: Float[Tensor, "grids feature_dim row column"]) -> Float[Tensor, ""]:
+    """
+    https://github.com/apchenstu/TensoRF/blob/4ec894dc1341a2201fe13ae428631b58458f105d/utils.py#L139
+
+    Args:
+        grids: stacks of explicit feature grids (stacked at dim 0)
+    Returns:
+        average total variation loss for neighbor rows and columns.
+    """
+    number_of_grids = grids.shape[0]
+    h_tv_count = grids[:, :, 1:, :].shape[1] * grids[:, :, 1:, :].shape[2] * grids[:, :, 1:, :].shape[3]
+    w_tv_count = grids[:, :, :, 1:].shape[1] * grids[:, :, :, 1:].shape[2] * grids[:, :, :, 1:].shape[3]
+    h_tv = torch.pow((grids[:, :, 1:, :] - grids[:, :, :-1, :]), 2).sum()
+    w_tv = torch.pow((grids[:, :, :, 1:] - grids[:, :, :, :-1]), 2).sum()
+    return 2 * (h_tv / h_tv_count + w_tv / w_tv_count) / number_of_grids
+
+
+class _GradientScaler(torch.autograd.Function):  # typing: ignore
+    """
+    Scale gradients by a constant factor.
+    """
+
+    @staticmethod
+    def forward(ctx, value, scaling):
+        ctx.save_for_backward(scaling)
+        return value, scaling
+
+    @staticmethod
+    def backward(ctx, output_grad, grad_scaling):
+        (scaling,) = ctx.saved_tensors
+        return output_grad * scaling, grad_scaling
+
+
+def scale_gradients_by_distance_squared(
+    field_outputs: Dict[FieldHeadNames, torch.Tensor],
+    ray_samples: RaySamples,
+) -> Dict[FieldHeadNames, torch.Tensor]:
+    """
+    Scale gradients by the ray distance to the pixel
+    as suggested in `Radiance Field Gradient Scaling for Unbiased Near-Camera Training` paper
+
+    Note: The scaling is applied on the interval of [0, 1] along the ray!
+
+    Example:
+        GradientLoss should be called right after obtaining the densities and colors from the field. ::
+            >>> field_outputs = scale_gradient_by_distance_squared(field_outputs, ray_samples)
+    """
+    out = {}
+    ray_dist = (ray_samples.frustums.starts + ray_samples.frustums.ends) / 2
+    scaling = torch.square(ray_dist).clamp(0, 1)
+    for key, value in field_outputs.items():
+        out[key], _ = cast(Tuple[Tensor, Tensor], _GradientScaler.apply(value, scaling))
+    return out
+
+
+def depth_ranking_loss(rendered_depth, gt_depth):
+    """
+    Depth ranking loss as described in the SparseNeRF paper
+    Assumes that the layout of the batch comes from a PairPixelSampler, so that adjacent samples in the gt_depth
+    and rendered_depth are from pixels with a radius of each other
+    """
+    m = 1e-4
+    if rendered_depth.shape[0] % 2 != 0:
+        # chop off one index
+        rendered_depth = rendered_depth[:-1, :]
+        gt_depth = gt_depth[:-1, :]
+    dpt_diff = gt_depth[::2, :] - gt_depth[1::2, :]
+    out_diff = rendered_depth[::2, :] - rendered_depth[1::2, :] + m
+    differing_signs = torch.sign(dpt_diff) != torch.sign(out_diff)
+    return torch.nanmean((out_diff[differing_signs] * torch.sign(out_diff[differing_signs])))
