@@ -1,4 +1,4 @@
-# Copyright 2022 The Nerfstudio Team. All rights reserved.
+# Copyright 2022 the Regents of the University of California, Nerfstudio Team and contributors. All rights reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -18,23 +18,28 @@ Generic Writer class
 from __future__ import annotations
 
 import enum
+import os
 from abc import abstractmethod
 from pathlib import Path
 from time import time
 from typing import Any, Dict, List, Optional, Union
 
 import torch
-import wandb
-from rich.console import Console
+from jaxtyping import Float
+from torch import Tensor
 from torch.utils.tensorboard import SummaryWriter
-from torchtyping import TensorType
 
 from nerfstudio.configs import base_config as cfg
 from nerfstudio.utils.decorators import check_main_thread, decorate_all
 from nerfstudio.utils.printing import human_format
+from nerfstudio.utils.rich_utils import CONSOLE
+
+
+def to8b(x):
+    """Converts a torch tensor to 8 bit"""
+    return (255 * torch.clamp(x, min=0, max=1)).to(torch.uint8)
+
 
-CONSOLE = Console(width=120)
-to8b = lambda x: (255 * torch.clamp(x, min=0, max=1)).to(torch.uint8)
 EVENT_WRITERS = []
 EVENT_STORAGE = []
 GLOBAL_BUFFER = {}
@@ -46,7 +51,6 @@ class EventName(enum.Enum):
 
     ITER_TRAIN_TIME = "Train Iter (time)"
     TOTAL_TRAIN_TIME = "Train Total (time)"
-    ITER_VIS_TIME = "Viewer Rendering (time)"
     ETA = "ETA (time)"
     TRAIN_RAYS_PER_SEC = "Train Rays / Sec"
     TEST_RAYS_PER_SEC = "Test Rays / Sec"
@@ -64,7 +68,7 @@ class EventType(enum.Enum):
 
 
 @check_main_thread
-def put_image(name, image: TensorType["H", "W", "C"], step: int):
+def put_image(name, image: Float[Tensor, "H W C"], step: int):
     """Setter function to place images into the queue to be written out
 
     Args:
@@ -145,7 +149,7 @@ def put_time(name: str, duration: float, step: int, avg_over_steps: bool = True,
         put_scalar(name, duration, step)
 
     if update_eta:
-        ## NOTE: eta should be called with avg train iteration time
+        # NOTE: eta should be called with avg train iteration time
         remain_iter = GLOBAL_BUFFER["max_iter"] - step
         remain_time = remain_iter * GLOBAL_BUFFER["events"][name]["avg"]
         put_scalar(EventName.ETA, remain_time, step)
@@ -180,25 +184,43 @@ def setup_local_writer(config: cfg.LoggingConfig, max_iter: int, banner_messages
     else:
         CONSOLE.log("disabled local writer")
 
-    ## configure all the global buffer basic information
+    # configure all the global buffer basic information
     GLOBAL_BUFFER["max_iter"] = max_iter
     GLOBAL_BUFFER["max_buffer_size"] = config.max_buffer_size
     GLOBAL_BUFFER["steps_per_log"] = config.steps_per_log
     GLOBAL_BUFFER["events"] = {}
 
 
-@check_main_thread
-def setup_event_writer(is_wandb_enabled: bool, is_tensorboard_enabled: bool, log_dir: Path) -> None:
-    """Initialization of all event writers specified in config
+def is_initialized():
+    """
+    Returns True after setup_local_writer was called
+    """
+    return "events" in GLOBAL_BUFFER
 
+
+@check_main_thread
+def setup_event_writer(
+    is_wandb_enabled: bool,
+    is_tensorboard_enabled: bool,
+    is_comet_enabled: bool,
+    log_dir: Path,
+    experiment_name: str,
+    project_name: str = "nerfstudio-project",
+) -> None:
+    """Initialization of all event writers specified in config
     Args:
         config: configuration to instantiate loggers
         max_iter: maximum number of train iterations
         banner_messages: list of messages to always display at bottom of screen
     """
     using_event_writer = False
+
+    if is_comet_enabled:
+        curr_writer = CometWriter(log_dir=log_dir, experiment_name=experiment_name, project_name=project_name)
+        EVENT_WRITERS.append(curr_writer)
+        using_event_writer = True
     if is_wandb_enabled:
-        curr_writer = WandbWriter(log_dir=log_dir)
+        curr_writer = WandbWriter(log_dir=log_dir, experiment_name=experiment_name, project_name=project_name)
         EVENT_WRITERS.append(curr_writer)
         using_event_writer = True
     if is_tensorboard_enabled:
@@ -208,7 +230,7 @@ def setup_event_writer(is_wandb_enabled: bool, is_tensorboard_enabled: bool, log
     if using_event_writer:
         string = f"logging events to: {log_dir}"
     else:
-        string = "Disabled tensorboard/wandb event writers"
+        string = "Disabled comet/tensorboard/wandb event writers"
     CONSOLE.print(f"[bold yellow]{string}")
 
 
@@ -216,7 +238,7 @@ class Writer:
     """Writer class"""
 
     @abstractmethod
-    def write_image(self, name: str, image: TensorType["H", "W", "C"], step: int) -> None:
+    def write_image(self, name: str, image: Float[Tensor, "H W C"], step: int) -> None:
         """method to write out image
 
         Args:
@@ -268,7 +290,7 @@ class TimeWriter:
     def __exit__(self, *args):
         self.duration = time() - self.start
         update_step = self.step is not None
-        if self.write:
+        if self.write and is_initialized():
             self.writer.put_time(
                 name=self.name,
                 duration=self.duration,
@@ -282,25 +304,36 @@ class TimeWriter:
 class WandbWriter(Writer):
     """WandDB Writer Class"""
 
-    def __init__(self, log_dir: Path):
-        wandb.init(project="nerfstudio-project", dir=str(log_dir), reinit=True)
+    def __init__(self, log_dir: Path, experiment_name: str, project_name: str = "nerfstudio-project"):
+        import wandb  # wandb is slow to import, so we only import it if we need it.
+
+        wandb.init(
+            project=os.environ.get("WANDB_PROJECT", project_name),
+            dir=os.environ.get("WANDB_DIR", str(log_dir)),
+            name=os.environ.get("WANDB_NAME", experiment_name),
+            reinit=True,
+        )
+
+    def write_image(self, name: str, image: Float[Tensor, "H W C"], step: int) -> None:
+        import wandb  # wandb is slow to import, so we only import it if we need it.
 
-    def write_image(self, name: str, image: TensorType["H", "W", "C"], step: int) -> None:
         image = torch.permute(image, (2, 0, 1))
         wandb.log({name: wandb.Image(image)}, step=step)
 
     def write_scalar(self, name: str, scalar: Union[float, torch.Tensor], step: int) -> None:
+        import wandb  # wandb is slow to import, so we only import it if we need it.
+
         wandb.log({name: scalar}, step=step)
 
     def write_config(self, name: str, config_dict: Dict[str, Any], step: int):
-        # pylint: disable=unused-argument
-        # pylint: disable=no-self-use
         """Function that writes out the config to wandb
 
         Args:
             config: config dictionary to write out
         """
-        wandb.config.update(config_dict)
+        import wandb  # wandb is slow to import, so we only import it if we need it.
+
+        wandb.config.update(config_dict, allow_val_change=True)
 
 
 @decorate_all([check_main_thread])
@@ -310,14 +343,14 @@ class TensorboardWriter(Writer):
     def __init__(self, log_dir: Path):
         self.tb_writer = SummaryWriter(log_dir=log_dir)
 
-    def write_image(self, name: str, image: TensorType["H", "W", "C"], step: int) -> None:
+    def write_image(self, name: str, image: Float[Tensor, "H W C"], step: int) -> None:
         image = to8b(image)
         self.tb_writer.add_image(name, image, step, dataformats="HWC")
 
     def write_scalar(self, name: str, scalar: Union[float, torch.Tensor], step: int) -> None:
         self.tb_writer.add_scalar(name, scalar, step)
 
-    def write_config(self, name: str, config_dict: Dict[str, Any], step: int):  # pylint: disable=unused-argument
+    def write_config(self, name: str, config_dict: Dict[str, Any], step: int):
         """Function that writes out the config to tensorboard
 
         Args:
@@ -326,6 +359,33 @@ class TensorboardWriter(Writer):
         self.tb_writer.add_text("config", str(config_dict))
 
 
+@decorate_all([check_main_thread])
+class CometWriter(Writer):
+    """Comet_ML Writer Class"""
+
+    def __init__(self, log_dir: Path, experiment_name: str, project_name: str = "nerfstudio-project"):
+        # comet_ml is slow to import, so we only do it if we need it.
+        import comet_ml
+
+        self.experiment = comet_ml.Experiment(project_name=project_name)
+        if experiment_name != "unnamed":
+            self.experiment.set_name(experiment_name)
+
+    def write_image(self, name: str, image: Float[Tensor, "H W C"], step: int) -> None:
+        self.experiment.log_image(image, name, step=step)
+
+    def write_scalar(self, name: str, scalar: Union[float, torch.Tensor], step: int) -> None:
+        self.experiment.log_metric(name, scalar, step)
+
+    def write_config(self, name: str, config_dict: Dict[str, Any], step: int):
+        """Function that writes out the config to Comet
+
+        Args:
+            config: config dictionary to write out
+        """
+        self.experiment.log_parameters(config_dict, step=step)
+
+
 def _cursorup(x: int):
     """utility tool to move the cursor up on the terminal
 
@@ -468,7 +528,7 @@ class LocalWriter:
 
             for i, mssg in enumerate(self.past_mssgs):
                 pad_len = len(max(self.past_mssgs, key=len))
-                style = "\x1b[6;30;42m" if self.banner_len and i >= len(self.past_mssgs) - self.banner_len + 1 else ""
+                style = "\x1b[30;42m" if self.banner_len and i >= len(self.past_mssgs) - self.banner_len + 1 else ""
                 print(f"{style}{mssg:{padding}<{pad_len}} \x1b[0m")
         else:
             print(curr_mssg)
