@@ -1,4 +1,3 @@
-# pylint: disable=protected-access
 """
 Default test to make sure train runs
 """
@@ -13,25 +12,31 @@ from nerfstudio.configs.method_configs import method_configs
 from nerfstudio.data.dataparsers.blender_dataparser import BlenderDataParserConfig
 from nerfstudio.data.dataparsers.minimal_dataparser import MinimalDataParserConfig
 from nerfstudio.engine.trainer import TrainerConfig
-from scripts.train import train_loop
+from nerfstudio.models.vanilla_nerf import VanillaModelConfig
+from nerfstudio.scripts.train import train_loop
 
 BLACKLIST = [
     "base",
     "semantic-nerfw",
     "instant-ngp",
     "instant-ngp-bounded",
-    "nerfacto",
-    "volinga",
+    "nerfacto-big",
     "phototourism",
     "depth-nerfacto",
-    "nerfplayer-ngp",
-    "nerfplayer-nerfacto",
+    "neus",
+    "generfacto",
+    "neus-facto",
+    "splatfacto",
 ]
 
 
-def set_reduced_config(config: TrainerConfig):
+def set_reduced_config(config: TrainerConfig, tmp_path: Path):
     """Reducing the config settings to speedup test"""
-    config.machine.num_gpus = 0
+    config.machine.device_type = "cpu"
+    if hasattr(config.pipeline.model, "implementation"):
+        setattr(config.pipeline.model, "implementation", "torch")
+    config.mixed_precision = False
+    config.use_grad_scaler = False
     config.max_num_iterations = 2
     # reduce dataset factors; set dataset to test
     config.pipeline.datamanager.dataparser = BlenderDataParserConfig(data=Path("tests/data/lego_test"))
@@ -44,21 +49,23 @@ def set_reduced_config(config: TrainerConfig):
 
     # reduce model factors
     if hasattr(config.pipeline.model, "num_coarse_samples"):
+        assert isinstance(config.pipeline.model, VanillaModelConfig)
         config.pipeline.model.num_coarse_samples = 4
     if hasattr(config.pipeline.model, "num_importance_samples"):
+        assert isinstance(config.pipeline.model, VanillaModelConfig)
         config.pipeline.model.num_importance_samples = 4
     # remove viewer
-    config.viewer.enable = False
+    config.viewer.quit_on_train_completion = True
 
-    # model specific config settings
-    if config.method_name == "instant-ngp":
-        config.pipeline.model.field_implementation = "torch"
+    # timestamp & output directory
+    config.set_timestamp()
+    config.output_dir = tmp_path / "outputs"
 
     return config
 
 
 @pytest.mark.filterwarnings("ignore::DeprecationWarning")
-def test_train():
+def test_train(tmp_path: Path):
     """test run train script works properly"""
     all_config_names = method_configs.keys()
     for config_name in all_config_names:
@@ -67,16 +74,16 @@ def test_train():
             continue
         print(f"testing run for: {config_name}")
         config = method_configs[config_name]
-        config = set_reduced_config(config)
+        config = set_reduced_config(config, tmp_path)
 
         train_loop(local_rank=0, world_size=0, config=config)
 
 
-def test_simple_io():
+def test_simple_io(tmp_path: Path):
     """test to check minimal data IO works correctly"""
     config = method_configs["vanilla-nerf"]
     config.pipeline.datamanager.dataparser = MinimalDataParserConfig(data=Path("tests/data/minimal_parser"))
-    config = set_reduced_config(config)
+    config = set_reduced_config(config, tmp_path)
     train_loop(local_rank=0, world_size=0, config=config)
 
 
